# -----------------------------------------------------------------------------
# $Id: Alias.pm 598 2004-09-28 01:44:59Z topia $
# -----------------------------------------------------------------------------
package Auto::Alias;
use strict;
use warnings;
use base qw(Module);
use Module::Use qw(Auto::AliasDB Auto::Utils);
use Auto::AliasDB;
use Auto::Utils;
use Mask;

sub new {
    my $class = shift;
    my $this = $class->SUPER::new(@_);
    Auto::AliasDB::setfile($this->config->alias,
			   $this->config->alias_encoding);
    $this;
}

sub message_arrived {
    my ($this,$msg,$sender) = @_;
    my @result = ($msg);

    if ($msg->command eq 'PRIVMSG') {

	if (Mask::match($this->config->confirm,$msg->param(1))) {
	    # οͤΥꥢpriv֤
	    my (undef,undef,$reply_as_priv,undef,undef)
		= Auto::Utils::generate_reply_closures($msg,$sender,\@result, 0); # Alias conversion disable.

	    my $alias = Auto::AliasDB->find_alias_prefix($msg->prefix);
	    if (defined $alias) {
		while (my ($key,$values) = each %$alias) {
		    map {
			$reply_as_priv->("$key: $_");
		    } @$values;
		}
	    }
	} else {
	    my (undef,undef,undef,$reply_anywhere,undef)
		= Auto::Utils::generate_reply_closures($msg,$sender,\@result, 1);

	    my $msg_from_modifier_p = sub {
		!defined $msg->prefix ||
		    Mask::match_deep([Mask::array_or_all($this->config->modifier('all'))],
				     $msg->prefix);
	    };

	    my ($temp) = $msg->param(1);
	    $temp =~ s/^\s*(.+)\s*$/$1/;
	    my ($keyword,$key,$value)
		= split(/\s+/, $temp, 3);

	    if (Mask::match($this->config->get('add'),$keyword)) {
		if ($msg_from_modifier_p->() && defined $key && defined $value) {
		    if (Auto::AliasDB->add_value_with_prefix($msg->prefix, $key, $value)) {
			$reply_anywhere->([$this->config->added_format('all')],
					  'key' => $key, 'value' => $value);
		    } else {
			$reply_anywhere->([$this->config->add_failed_format('all')],
					  'key' => $key, 'value' => $value);
		    }
		}
	    } elsif (Mask::match($this->config->get('remove'),$keyword)) {
		if ($msg_from_modifier_p->() && defined $key) {
		    my $count = Auto::AliasDB->del_value_with_prefix($msg->prefix, $key, $value);
		    if ($count) {
			$reply_anywhere->([$this->config->removed_format('all')], 'key' => $key, 'value' => $value, 'count' => $count);
		    } else {
			$reply_anywhere->([$this->config->remove_failed_format('all')], 'key' => $key, 'value' => $value);
		    }
		}
	    }
	}
    }
    return @result;
}

1;

=pod
info: 桼ꥢδԤʤޤ
default: off

# ꥢϴŪname,userĤΥեɤäƤꡢ
# 줾桼̾桼ޥɽޤ

# ꥢեΥѥȡΥ󥳡ǥ󥰡
# ΥեϼΤ褦ʥեޥåȤǤ롣
# 1. 줾ιԤϡ<>: <>פηǤ롣
# 2. ιԤǡƥ桼ڤ롣
# 3. <>ϥޤǶڤʣͤȤ롣
#
# ꥢե:
#
# name: sample
# user: *!*sample@*.sample.net
#
# name: sample2,[sample2]
# user: *!sample2@*.sample.net,*!sample2@*.sample2.net
#
alias: alias.txt
alias-encoding: euc

# ȯ򤷤ͤΥꥢϿƤСpriv롣
confirm: ꥢǧ

# <addǻꤷ> user *!*user@*.user.netפΤ褦ˤƾɲá
# ȯ򤷤ͤΥꥢ̤ϿäϡuserΤ߼դƿɲäȤ롣
add: ꥢɲ

# <removeǻꤷ> name 桼פΤ褦ˤƾ
# userƺ줿ꥢ¾ξ(name)ޤƾǤ롣
remove: ꥢ

# åɲä줿Ȥȿꤷޤ
# ʥåȯݤΥեޥåȤꤷޤ
# ꥢִͭǤ#(nick.now)#(channel)
# 줾nickִͥ̾ޤ
# #(key)#(value)ϡɲä줿ִͤޤ
added-format: #(name|nick.now): ꥢ #(key)  #(value) ɲäޤ
add-failed-format: #(name|nick.now): ꥢ #(key) ɲä˼Ԥޤ

# å줿Ȥȿꤷޤ
# added-formatǻǤΤƱǤ
removed-format: #(name|nick.now): ꥢ #(key)  #(value) ޤ
remove-failed-format: #(name|nick.now): ꥢ #(key) κ˼Ԥޤ

# ꥢɲäƤ͡ά줿ϡ*!*@*פȸ롣
modifier: *!*@*
=cut
